/*
 * calmwm - the calm window manager
 *
 * Copyright (c) 2004 Marius Aamodt Eriksen <marius@monkey.org>
 * All rights reserved.
 *
 * $Id: conf.c,v 1.20 2004/07/10 20:29:51 provos Exp $
 */

#include "headers.h"
#include "calmwm.h"

#ifndef timespeccmp
#define timespeccmp(tsp, usp, cmp)                                      \
        (((tsp)->tv_sec == (usp)->tv_sec) ?                             \
            ((tsp)->tv_nsec cmp (usp)->tv_nsec) :                       \
            ((tsp)->tv_sec cmp (usp)->tv_sec))
#endif

#define CONF_MAX_WINTITLE 256

#define CONF_IGNORECASE   0x01

/*
 * Match a window.
 */
struct winmatch {
	TAILQ_ENTRY(winmatch) entry;

	char title[CONF_MAX_WINTITLE];
	int  opts;
};

TAILQ_HEAD(winmatch_q, winmatch);
struct winmatch_q ignoreq;

/* XXX - until we get a real configuration parser. */
#define WINMATCH_ADD(queue, str) do {			\
	struct winmatch *wm;				\
	XCALLOC(wm);					\
	strlcpy(wm->title, str, sizeof(wm->title));	\
	wm->opts |= CONF_IGNORECASE;			\
	TAILQ_INSERT_TAIL(queue, wm, entry);		\
} while (0)

#define KB_ADD(c, which, sym, mask, flgs) do {	\
	c->kb[which].modmask = mask;		\
	c->kb[which].keysym = sym;		\
	c->kb[which].flags = flgs;		\
} while (0)

/* #define SYSTR_PRE "systrace -C -g /usr/local/bin/notification -d /usr/home/marius/policy/X11 " */

/* Initializes the command menu */

void
conf_cmd_init(struct conf *c)
{
	TAILQ_INIT(&c->cmdq);
}

/* Removes all static entries */

void
conf_cmd_clear(struct conf *c)
{
	struct cmd *cmd, *next;

	for (cmd = TAILQ_FIRST(&c->cmdq); cmd != NULL; cmd = next) {
		next = TAILQ_NEXT(cmd, entry);

		/* Do not remove static entries */
		if (cmd->flags & CMD_STATIC)
			continue;

		TAILQ_REMOVE(&c->cmdq, cmd, entry);
		free(cmd);
	}
}

/* Add an command menu entry to the end of the menu */

void
conf_cmd_add(struct conf *c, char *image, char *label, int flags)
{
	struct cmd *cmd;
	XMALLOC(cmd);
	cmd->flags = flags;
	strlcpy(cmd->image, image, sizeof(cmd->image));
	strlcpy(cmd->label, label, sizeof(cmd->label));
	TAILQ_INSERT_TAIL(&c->cmdq, cmd, entry);
}

int
conf_cmd_changed(char *path)
{
#ifdef __OpenBSD__
	static struct timespec old_ts;
#else
	static time_t old_time;
#endif
	struct stat sb;
	int changed;

	/* If the directory does not exist we pretend that nothing changed */
	if (stat(path, &sb) == -1 || !(sb.st_mode & S_IFDIR))
		return (0);

#ifdef __OpenBSD__
	changed = !timespeccmp(&sb.st_mtimespec, &old_ts, ==);
	old_ts = sb.st_mtimespec;
#else
	changed = old_time != sb.st_mtime;
	old_time = sb.st_mtime;
#endif


	return (changed);
}

void
conf_cmd_populate(struct conf *c, char *path)
{
	DIR *dir;
	struct dirent *file;
	char fullname[PATH_MAX];
	int off;

	if (strlen(path) >= sizeof (fullname) - 2)
		errx(1, __FUNCTION__": directory name too long");

	dir = opendir(path);
	if (dir == NULL)
		err(1, "opendir");

	strlcpy(fullname, path, sizeof (fullname));
	off = strlen(fullname);
	if (fullname[off - 1] != '/') {
		strlcat(fullname, "/", sizeof(fullname));
		off++;
	}

	while ((file = readdir(dir)) != NULL) {
		char *filename = file->d_name;
		if (!strcmp(filename, ".") || !strcmp(filename, ".."))
			continue;

		strlcpy(fullname + off, filename, sizeof(fullname) - off);

		/* Add a dynamic entry to the command menu */
		conf_cmd_add(c, fullname, filename, 0);
	}
}

void
conf_setup(struct conf *c)
{
	char *home = getenv("HOME");
	if (home == NULL)
		errx(1, "No HOME directory.");
	snprintf(c->menu_path, sizeof(c->menu_path), "%s/.calmwm", home);

	c->flags = 0;
	
	TAILQ_INIT(&ignoreq);
	WINMATCH_ADD(&ignoreq, "XMMS");
	WINMATCH_ADD(&ignoreq, "XMMS Playlist");
	WINMATCH_ADD(&ignoreq, "xwi");
	WINMATCH_ADD(&ignoreq, "xapm");
	WINMATCH_ADD(&ignoreq, "xclock");

/* 	KB_ADD(c, KB_DELETE, XK_Escape, ControlMask|Mod1Mask, KBFLAG_NEEDCLIENT); */
	KB_ADD(c, KB_NEWTERM0, XK_Return, ControlMask|Mod1Mask, 0);
	KB_ADD(c, KB_NEWTERM1, XK_Delete, ControlMask|Mod1Mask, 0);
	KB_ADD(c, KB_HIDE, XK_Return, Mod1Mask, KBFLAG_NEEDCLIENT);
	KB_ADD(c, KB_LOWER, XK_Down, Mod1Mask, KBFLAG_NEEDCLIENT);
	KB_ADD(c, KB_RAISE, XK_Up, Mod1Mask, KBFLAG_NEEDCLIENT);
	KB_ADD(c, KB_SEARCH, XK_slash, Mod1Mask, 0);
	KB_ADD(c, KB_CYCLE, XK_Tab, Mod1Mask, KBFLAG_FINDCLIENT);
	KB_ADD(c, KB_LABEL, XK_l, ControlMask|Mod1Mask, KBFLAG_NEEDCLIENT);

	KB_ADD(c, KB_DELETE, XK_x, ControlMask|Mod1Mask, KBFLAG_NEEDCLIENT);
	KB_ADD(c, KB_GROUPSELECT, XK_Escape, ControlMask|Mod1Mask, 0);
	KB_ADD(c, KB_GROUP_1, XK_1, ControlMask|Mod1Mask, 0);
	KB_ADD(c, KB_GROUP_2, XK_2, ControlMask|Mod1Mask, 0);
	KB_ADD(c, KB_GROUP_3, XK_3, ControlMask|Mod1Mask, 0);
	KB_ADD(c, KB_GROUP_4, XK_4, ControlMask|Mod1Mask, 0);
	KB_ADD(c, KB_GROUP_5, XK_5, ControlMask|Mod1Mask, 0);
	KB_ADD(c, KB_GROUP_6, XK_6, ControlMask|Mod1Mask, 0);
	KB_ADD(c, KB_GROUP_7, XK_7, ControlMask|Mod1Mask, 0);
	KB_ADD(c, KB_GROUP_8, XK_8, ControlMask|Mod1Mask, 0);
	KB_ADD(c, KB_GROUP_9, XK_9, ControlMask|Mod1Mask, 0);
	KB_ADD(c, KB_NOGROUP, XK_0, ControlMask|Mod1Mask, 0);
	KB_ADD(c, KB_NEXTGROUP, XK_Right, ControlMask|Mod1Mask, 0);
	KB_ADD(c, KB_PREVGROUP, XK_Left, ControlMask|Mod1Mask, 0);

	conf_cmd_init(c);

	c->term0 = "xterm";
	c->term1 = "xterm";
}

int
conf_get_int(struct client_ctx *cc, enum conftype ctype)
{
	int val = -1, ignore = 0;
	char *wname;
	struct winmatch *wm;

	wname = cc->name;

	/* Can wname be NULL? */

	if (wname != NULL) {
		TAILQ_FOREACH(wm, &ignoreq, entry) {
			if (((wm->opts & CONF_IGNORECASE) &&
				strcasecmp(wm->title, wname) == 0) ||
			    (!(wm->opts & CONF_IGNORECASE) &&
				strcmp(wm->title, wname) == 0)) {
				ignore = 1;
				break;
			}
		}
	} else
		ignore = 1;

	switch (ctype) {
	case CONF_BWIDTH:
		/* XXX this will be a list, specified in the
		 * configuration file.
		 */
		val = ignore ? 0 : 3;
		break;
	case CONF_IGNORE:
		val = ignore;
		break;
	}

	return (val);
}

void
conf_client(struct client_ctx *cc)
{
	cc->bwidth = conf_get_int(cc, CONF_BWIDTH);
	cc->flags |= conf_get_int(cc, CONF_IGNORE) ? CLIENT_IGNORE : 0;
}
